/* -*- compile-command: "env CC=gcc mcc -Wall -g jmath.c -o jmath -lreadline -ltermcap -lsocket -lnsl" -*- */
/* Copyright (C) 1998,99 Jim Radford

 * This File is JMath

 * JMath is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 
 * JMath is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 
 * You should have received a copy of the GNU General Public License
 * along with JMath; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/* What is cool:
 *  o Full emacs command line editing and history (via readline)
 *  o Symbol completion
 *  o Filename completion within strings and after << and >>
 *  o Automatic resizing of of display
 *  o Moves cursor to position of syntax errors
 * What would be cool:
 *  o input also accepted on a fifo (without completion)
 *  o indentation for output
 *  o small output on same line with output prompt (see questions)
 *  o Completion
 *     - Commands after !
 *     - Absolute filenames
 *     - Completion using $Path
 *  o To get rid of the order n^2 processing of long input lines we might
 *     - $SyntaxHandler = $JMath`SyntaxHandler[#2] &
 *        This way we don't send the whole string unless we need to.
 *        We can return $Failed (on double return) and a new string on
 *        an edit.  I'm not sure if it will work, but it might.
 *  o Set up a Message Handler
 * Bugs
 *  o We don't clean up properly after SyntaxLength is wrong.
 *  o We shouldn't print banner unless asked.
 *  o We don't accept an alternate name for "math"
 *  o Can't complete on "<< /local/m" ("<< /loca" and "<< "/local/m" work)
 *  o C-c during Plot crashes (math too)
 *  o if ?usage:: C-d gives Syntax Packet (which I consider an error)
 *  o SyntaxLength["?J::"] != 2  (math bug) (SyntaxQ wrong also)
 *  o ?OutputFor,   gives syntax error
 *  o ?/: give syntax error
 *  o a /: a ^:= a gives syntax error
 *  o completion won't close double quotes if readline < 2.2
 *  o Completing JMa gives "JMath` "; how can we stop ' ' after '`'
 *     - We could have all completions in all contexts show up
 *     - We could hack rl_completion_append_character temporarily
 *  o Because of some readline interaction, we now do
 *    the window-change stuff for EVERY command.
 * Questions for Mr. MathLink:
 *  o How can I get the prompt in the output without $PrePrint
 *  o C-c during graphics kills math (regular math-3 and hence jmath)
 *  o How to clean up on a SIGTERM given that we could be "interupted"
 *  o Why do Run["ls"] and !ls produce no output?  What about stdin?
 *     - Ok, I fixed this by duping stdout to std{in,out}. :-(
 *  o Why does MLClose() always return error 999?
 */

#define JMATH_VERSION "0.9.10"

/* A little help with MathLink since the doccumentation is not complete:
 * The link stream consists of a series of function with args. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <termios.h> /* struct winsize */
#include "mathlink.h"
#include <readline/readline.h>
#include <readline/history.h>

extern char *xmalloc(); /* readine has our copy */

static int actually_use_line; /* Save line in history but don't use it */

static char *
rl_gets (const char *prompt)
{
  static char *line_read = (char *)NULL;

  do {
    actually_use_line=1;

    if (line_read) {
      free (line_read);
      line_read = (char *)NULL;
    }
    line_read = readline ((char*)prompt);
    if (line_read && *line_read)
      add_history (line_read);
    
  } while(!actually_use_line);
  
  return (line_read);
}

static MLINK   lp;

static void read_and_print_expression(MLINK lp,FILE *out);
static void read_and_print_unquoted_string(MLINK lp,FILE *out,
					   char *extra_string);

static void DebugShowNextPacket(MLINK lp)
{
#if 0 /* main debug switch */
  MLMARK mark = MLCreateMark(lp);
  if(mark) { /* I can't seem to call this the first time */
//    do {
  read_and_print_expression(lp,stderr);
//    } while(MLReady(lp));
    MLSeekMark(lp,mark,0);
    MLDestroyMark(lp,mark);
  }
#endif
}
static int DebugMLNewPacket(MLINK lp)
{
  DebugShowNextPacket(lp);
  return MLNewPacket(lp);
}
static int DebugMLNextPacket(MLINK lp)
{
  DebugShowNextPacket(lp);
  return MLNextPacket(lp);
}

static void get_message();

static void
ml_expect(int packet,int type)
{
  int ans;

  while((ans = DebugMLNextPacket(lp)) == MESSAGEPKT)
    get_message();

  if(ans != packet)
    fprintf(stderr,"Wrong answer (%d) expexted %d!\n",ans,packet);

  if((ans=MLGetNext(lp)) != type)
    fprintf(stderr,"Wrong Type '%c' expected '%c'!\n",(char)ans,(char)type);
}

static void
get_message()
{
  MLNewPacket(lp); /* Dropping Message Symbol */
  ml_expect(TEXTPKT,MLTKSTR);
  read_and_print_unquoted_string(lp,stderr,"\n");
  rl_on_new_line();
}


/* the returned strings must be allocated with malloc and are 
   freed by readline. 
*/
static char *
symbol_completion_entry_function(char *text,int state)
{
  static long_st count=0;
  const char *tmp;
  char *match;

  if(!state) {
    char *wild_carded;
/*
    fprintf(stderr,"\ncomplete: '%s'\n",text); rl_on_new_line();
*/
/*     if(strchr(rl_special_prefixes,text[0])) text++; */
    wild_carded = (char*)xmalloc(strlen(text)+2);
    strcat(strcpy(wild_carded,text),"*");
    MLPutFunction(lp,"JMath`Private`" "NamesAndContexts",1L);
    MLPutNext(lp,MLTKSTR);
    MLPutByteString(lp,wild_carded,strlen(wild_carded));
    MLEndPacket(lp);
    free(wild_carded);
    ml_expect(RETURNPKT,MLTKFUNC);
    if(!MLCheckFunction(lp,"List",&count)) {
      fprintf(stderr,"\nWere is my List!!!\n");
      return NULL;
    }
  }

  if(state == count) {
/*
    fprintf(stderr,"Done\n"); rl_on_new_line();
*/
    count=0;
    return NULL;
  }

  MLGetString(lp,&tmp); /* change this to MLGetByteString */
  match = (char *)xmalloc(strlen(tmp)+1);
  strcpy(match,tmp);
  MLDisownString(lp,tmp);

/*
  fprintf(stderr,"%d) %s\n",state,match);
*/

  return match;
}

static int 
boring(char *filename)
{
  static char *boring_extensions[] = {"~",".o"};
  int i;
  for(i=0; i<sizeof(boring_extensions)/sizeof(*boring_extensions); i++) {
    int f=strlen(filename), b=strlen(boring_extensions[i]);
    if(f >= b && !strcmp(filename + f - b, boring_extensions[i]))
      return 1;
  }
  return 0;
}

static char **
jmath_ignore_boring_files(char **matches)
{
  char **good,**match;
  if(matches) {
    for(match=good=matches; *match; match++)
      if(boring(*match))
        free(*match);
      else
        *good++ = *match;
    *good=0;
  }
  return matches;
}

static char **
jmath_attempt_filename_entry_function(char *text, int start, int end)
{
#ifndef NoFC
  int indicator = start - 1;

/*   fprintf(stderr,"\njmath completer: (%d,%d) \"%s\"\n",start,end,text); */

  if(end == 0 || indicator < 0) {
/*     rl_forced_update_display (); */
    return NULL;
  }

  while(indicator >= 0 && isspace((int)rl_line_buffer[indicator]))
    indicator--;

/*    fprintf(stderr,"indicator %d '%c'\n", */
/*            indicator,rl_line_buffer[indicator]); */

  switch(rl_line_buffer[indicator]) {
  case '>': 
  case '<':
    if(indicator > 0 && /* check for << and >> */
       rl_line_buffer[indicator-1] == rl_line_buffer[indicator]) {
      rl_attempted_completion_over = 1;
/*       rl_forced_update_display (); */
      return jmath_ignore_boring_files
        (completion_matches(text,filename_completion_function));
    }
    break;
  case '"':
    {
      int instring = 1;
      while(--indicator >= 0)
        if(rl_line_buffer[indicator] == '"')
          if(indicator == 0 || rl_line_buffer[indicator-1] != '\\')
            instring ^= 1;
      if(!instring)
        break;
    }
    rl_attempted_completion_over = 1;
/*     rl_forced_update_display (); */
    return jmath_ignore_boring_files
      (completion_matches(text,filename_completion_function));
  case '!':
  case '[':
  default:
/*     rl_forced_update_display (); */
    return NULL;
  }
#endif
/*   rl_forced_update_display (); */
  return (char **)NULL;
}

#if 0
/* return /(<<||>>)\s*[\w`/.\\!-_:$*~?]+\p/ where \p matches the point. */
int looking_at_unquoted_filename_char_p(char *line,int point)
{ 
  fprintf(stderr,"\nLooking for an unquoted filename.\n");
  while(point > 0  &&
        (isalnum(rl_line_buffer[point]) || 
         strchr("`/.\\!-_:$*~?",rl_line_buffer[point])))
    point--;
  while(point > 0 &&
        isspace(rl_line_buffer[point]))
    point--;
  if(point == 0 || 
     strcmp("<<",&rl_line_buffer[point-1]) || 
     strcmp(">>",&rl_line_buffer[point-1]))
    return 0;
  else {
    fprintf(stderr,"\nWe found an unquoted filename at %d\n",point);
    rl_forced_update_display();
    return 1;
  }
}
#endif

/* This is here just so it is next to the actual completion function */
static void
setup_readline_filename_completion()
{
#ifndef NoFC
  /* In general we assume that any quoted string is a filename for the
     sake of completion.  It may contain any characters.  In addition
     we allow unquoted filenames after <<, >> and >>> as long as they
     contain only alphanumeric characters, special characters, or any
     of "`/.\\!-_:$*~?" or things match nested (\[[^ \t\n]+\])+. See
     981 of "The Mathematica Book" (third edition).  Special
     characters are of the form \[EatMe]; we ignore unquoted versions
     of these too for now.

     We ignore the fact that any character except spaces tabs and
     newlines may appear in unquoted filenames if they appear in
     nested square brackets. 

  */

  rl_completer_quote_characters = "\"";
  rl_basic_quote_characters = "\"";
  /* These are the valid unquoted filename characters "!*_-?.~/" */
  rl_filename_quote_characters =
    "\"@#%^&()+{}|=[]'<>,"  /* These NEED to be quoted for mathematica */
    "!*_-?.~/"              /* These NEED to be quoted for readline since
                               they are word break characters */
    /* Well it turns out that stupid readline doesn't check to
       see if "/" needs to be quoted before inserting it as the
       directory separator, so we force all filenames to be quoted,
       otherwise we have problems as "/" turns into a non word
       break character when in a filename.  */
    ".-_/"                  
    "abcdefghijklmnopqrstuvwxyz"
    "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
  /*rl_special_prefixes="/.-~";*//* Leave these word separators on filenames */
  rl_attempted_completion_function = jmath_attempt_filename_entry_function;
  /* This doesn't actually get called unless
     rl_completion_entry_function is filename_completion_function
     which it is not.  It should be called if
     filename_completion_function ever returns non-NULL during a
     completion cycle.  I should send in a bug report about this.  So
     we leave this here to remind ourselves. */
  rl_ignore_some_completions_function = (Function*)jmath_ignore_boring_files;
#endif
}

static int /* This is now readline (4.0) */
delete_char_or_list_or_eof(int count, int key)
{
  if (rl_point == rl_end)
    return rl_possible_completions(count,key);
  else
    return rl_delete(count,key);
}

/* History and editing can only happen on the same line (readline
 * restriction I think) so we can't have embedded newlines like the
 * default math.  We will however emulate the behavior that it allows
 * them, we just wont include them
 */

static int
math_return_key(int count/*ignored*/, int key)
{
/*  extern int rl_pending_input; */
  static int syntax_length=-1;

  int old_syntax_length = syntax_length;
  /*static int x=0;fprintf(stderr,"return (%d) [%x] %d\n",count,key,++x); */

  /* Why did I do this??? */
  if(rl_end == 0 || *rl_line_buffer == '!') { /* Should do shell completion */
    rl_newline(count,key); /*    rl_pending_input='\n'; */
    syntax_length = -1;
    return 0;
  }

  /* Backslash at the end of a line forces a line continuation so we
     just delete it */
  if(rl_end > 0 && rl_line_buffer[rl_end-1] == '\\') {
    rl_line_buffer[--rl_end]='\0';
    if(rl_point > rl_end) 
      rl_point = rl_end;
    return 0;
  }

  MLPutFunction(lp,"SyntaxLength",1L);
  MLPutNext(lp,MLTKSTR);
  MLPutByteString(lp,rl_line_buffer,rl_end);
  MLEndPacket(lp);

  ml_expect(RETURNPKT,MLTKINT);
  MLGetInteger(lp,&syntax_length);

  if(syntax_length == rl_end) {
    /*    fprintf(stderr,"Complete Expression %d\n",syntax_length); */
    syntax_length = -1;
    rl_newline(count,key); /*    rl_pending_input='\n'; */
  }
  else if(syntax_length == old_syntax_length) {
    /* fprintf(stderr,"Dropping Incomplete Expression %d\n",syntax_length); */
    /* This technique is not really sound.  It should give you another
       chance if you acutually tried to fix the line, but I don't know
       an easy way to sense that. */
    /*    rl_message("Gone!"); sleep(1); rl_clear_message(); */
    
    /* Can't just add_history(rl_line_buffer) because it confuses
       readline().  We must wait and then add it. */
    actually_use_line=0;
    rl_newline(count,key); /*    rl_pending_input='\n'; */
    syntax_length=-1;
  }
  else  if(syntax_length > rl_end) {
    /*fprintf(stderr,"Incomplete Expression %d, Waiting for more.\n",syntax_length);*/
    /* It seems to return this if the expression is just incomplete    */
    /* Since we can't go to the next line (readline) we'll just ignore */
    if(syntax_length-rl_end != 2) {
      rl_message("\nsyntax? %-10d",syntax_length-rl_end);
      sleep(1);
      rl_clear_message();
    }
  }
 else if(syntax_length < rl_end)
    rl_point = syntax_length;
  return 0;
}

static int evaluating = 0;
static int interrupt = 0;

#include <signal.h>

/* You will not get a respone from the kernel until it is evaluating,
   so don't waste our time.
   */
static void 
user_interrupt_function()
{
  signal(SIGINT,user_interrupt_function);
  if(evaluating) {
    fprintf(stderr,"Interrupting\n");
    MLPutMessage(lp,MLInterruptMessage);
  } else {
    fprintf(stderr,"Interrupting myself if possible\n");
    interrupt = 1;
  }
  evaluating=0;
}

static void error();
static MLEnvironment ml_env = 0;

/* I'm not sure what will kill the math process in all cases.
 * Specifically I don't know how to kill the math process if
 * we are in the middle of a hanging MLNextPacket() so I try
 * a few things (which don't work).
 */
static void 
termination_interrupt_function()
{
  signal(SIGTERM,termination_interrupt_function);
  fprintf(stderr,"About to termintate\n");
  MLPutFunction(lp, "Exit", 0L);
  MLEndPacket(lp);
  MLPutMessage(lp,MLInterruptMessage);
  MLPutString(lp,"exit");
  MLEndPacket(lp);
  MLClose(lp);
  MLDeinitialize(ml_env);
  exit(1); /* Maybe you can't do this */
}

static int new_page_width = 0;

static void
window_changed_interrupt_function()
{
  struct winsize ws;
  signal(SIGWINCH,window_changed_interrupt_function);
  if(!ioctl(1,TIOCGWINSZ,&ws) && ws.ws_col) {
/*    fprintf(stderr,"\nRecalculate Width (%d)\n",ws.ws_col); */
    new_page_width = ws.ws_col - 2;
  }
}

static void
set_new_page_width()
{
  MLPutFunction(lp,"SetOptions",2L);
  MLPutSymbol(lp,"$Output");
  MLPutFunction(lp,"Rule",2L);
  MLPutSymbol(lp,"PageWidth");
  MLPutInteger(lp,new_page_width);
  MLEndPacket(lp);
  new_page_width = 0;
  ml_expect(RETURNPKT, MLTKFUNC);
  MLNewPacket(lp); /* Drop Returned List of Options */
}

int main(int argc, char *argv[])
{
  int i;
  long error;
#define linkname_and_argv nargv[3]
  char *nargv[] = {"-linkmode", "launch", 
                   "-linkname", 0/*linkname[]+argv[]*/, 
                   0/*Termination*/};
  /* We pass only -linkname "linkname" to MLOpenArgv. linkname is the
     name of the program we want to call.  For us that is math with a
     bunch of args.  We have to quote things for the shell.  We also
     have to save and restore $ParentLink since files like init.m and
     Motif.m wont initialize normal "text mode" graphics when
     $ParentLink=!=Null.  We also quote (which isn't quite right if
     they already have quotes) the args passed to us and tack them on
     to the linkname too. */
  char linkname[] = 
    "math -mathlink"
      " -runfirst 'JMath`$JMathVersion=\"" JMATH_VERSION "\";'"
      /* Motif.m only gets loaded if $ParentLink === Null (which it doesn't) */
      " -run '($Output={};If[Environment[\"DISPLAY\"]=!=$Failed,"
                             "<<Motif.m,<<Terminal.m];$Output=#;)& @ $Output;' "
      /* A bit too tricky, but I couldn't get CallPacket to work */
      " -run 'JMath`Run = System`Run[##,\"1>&2 0<&2\"]&;'"
      /* This is a utility function used by the completion routines */
      " -run 'JMath`Private`NamesAndContexts[JMath`Private`n_String]:="
      "  Select[Contexts[],StringMatchQ[#,JMath`Private`n]&] ~Join~ Names[JMath`Private`n];'"
      /* I don't like this, but I can't see another way to get the default
         Run to be mine which then calls System`Run. */
      " -run '$ContextPath = {\"Global`\", \"JMath`\", \"System`\"};'";

  int arg_string = sizeof(linkname);

  for(i=1; i<argc; i++)
    arg_string += strlen(argv[i]) + 4; /* 4 is for a space, two 's, and '\0' */

  linkname_and_argv = strcpy(xmalloc(arg_string),linkname);
  for(i=1; i<argc; i++) {
    strcat(linkname_and_argv, " '");
    strcat(linkname_and_argv, argv[i]);
    strcat(linkname_and_argv, "'");
  }

/*   fprintf(stderr,"starting mathematica with: %s\n",linkname_and_argv);  */

  if((ml_env = MLInitialize(NULL)) == NULL) return 1;
  if((lp     = MLOpenArgv(ml_env, nargv, nargv+sizeof(nargv)/sizeof(*nargv)-1, 
                          &error))   == NULL) return 2;
  
  printf(
"JMath %s, a FrontEnd for Mathematica, copyright 1996-99 Jim Radford\n"
"Readline %s portions copyright 1987-99 Free Software Foundation, Inc.\n"
"MathLink portions copyright 1986-96 Wolfram Research, Inc.\n"
,JMATH_VERSION,rl_library_version);

  if(!MLActivate(lp)) { /* Wait for a connection */
    fprintf(stderr, "Couldn't connect to Mathematica\n");
    return 3;
  }

  rl_readline_name = "JMath";
  /* For some reason I get both '\n' and '\r'.  Which is random */
  rl_bind_key('\r',math_return_key);
  rl_bind_key('\n',math_return_key);
  rl_bind_key('\004',delete_char_or_list_or_eof); /* Like tcsh */
  rl_basic_word_break_characters = " \t\n\"!@#%^&*()_+{}|-=[]'?<>,./~";
  rl_completion_entry_function = (Function*)symbol_completion_entry_function;
  setup_readline_filename_completion();

  /* Maybe use signal and put signal in handlers to support SunOS */
  signal(SIGINT,user_interrupt_function);
  window_changed_interrupt_function(); /* kickstart for Linux */
  signal(SIGTERM,termination_interrupt_function);

  while(1)
    {
      int packet;
      char *line="";

      if(interrupt) {
        fprintf(stderr,"Didn't get a chance to interrupt myself.\n");
      }
      interrupt = 0;

      switch(packet = DebugMLNextPacket( lp)) {
      case INPUTNAMEPKT :  /* Regular In[n]:= */
      case INPUTPKT :      /* Input["prompt"], maybe this shoudn't use
			      syntax checking */

	/* 
	 * This drains any return packets, that we obviously
	 * didn't care about since they weren't already dealt
	 * with, so they don't conflict with the return packet of
	 * subsequent commands. these might have come from 
	 * commands that occured while an IN was pending so
	 * that you couldn't get the return without getting the IN.
	 */
	/*
	  while(  MLNextPacket( lp) == RETURNPKT ) {
	  fprintf(stderr,"Ignoring some return packet.\n");
	  MLNewPacket( lp);
	  if( MLError( lp)) error( lp);
	  }
	  */
	/*printf("Input %s Packet\n",packet==INPUTNAMEPKT?"Name":"");*/
	{
	  const char *prompt;
	  MLGetString(lp,&prompt);
	  evaluating = 0; /* No output on empty line, so must reset here*/
	  if((line = rl_gets(prompt))) {
		
	    if(new_page_width)      /* not in the signal so it's */
	      set_new_page_width(); /* done in known sequence    */
		
	    MLPutFunction( lp, "EnterTextPacket", 1L);
/* 	    MLPutFunction( lp, "EnterExpressionPacket", 1L); */
	    MLPutNext(lp,MLTKSTR);
	    MLPutByteString(lp,line,strlen(line));
	    MLEndPacket(lp);
	    evaluating = 1;
	  } else {
            MLPutFunction(lp, "Exit", 0L);
            MLEndPacket(lp);
          }
	  MLDisownString(lp,prompt);
	}
      break;
      case OUTPUTNAMEPKT :
	evaluating = 0;
	read_and_print_unquoted_string(lp,stdout,"\n");
	break;
      case TEXTPKT :
	read_and_print_unquoted_string(lp,stdout,"\n");
	if(!evaluating) /* Do we need this? */
	  rl_on_new_line();
	break;
      case RETURNTEXTPKT :
	evaluating = 0;
	read_and_print_unquoted_string(lp,stdout,"\n");
	break;
      case ILLEGALPKT :
	fprintf(stderr,"illegal Packet!!\n");
	if(MLError(lp) != MLEOK) {
	  fprintf(stderr,"Jim Says: %s\n",MLErrorMessage(lp));
	  if(!MLClearError(lp)) {
	    fprintf(stderr,"Can't clear error\n");
	    line = 0; /*break out of while*/
	  }
	}
	DebugMLNewPacket(lp); /* Dropping what ?? */
	break;
      case SUSPENDPKT : /* Quit ir Exit */
/* 	printf("Bye from Jim's Mathematica\n"); */
/* 	exit(0); */
        line = 0;
	break;
      case MESSAGEPKT :
	MLNewPacket(lp);  /* Drop Message Symbol */
	break;
      case RETURNPKT :
	fprintf(stderr,"Return packet I wasn't expecting.\n");
	MLNewPacket(lp);  /* Drop it */
	break;
      case MENUPKT : /* There is still Some problems with multiple ^Cs */
	{
	  int menu=0;
	  const char *prompt=NULL;
	  evaluating = 0;
	  fprintf(stderr,"\n");
	  MLGetInteger(lp,&menu);
	  MLGetString(lp,&prompt);
	  if(!menu) {
	    ml_expect(TEXTPKT,MLTKSTR);
	    read_and_print_unquoted_string(lp,stdout,"\n");
	  }
	  rl_bind_key('\r',rl_newline);
	  rl_bind_key('\n',rl_newline);
	  MLPutNext(lp,MLTKSTR);
	  if((line=readline((char*)prompt))) {
	    MLPutByteString(lp,line,strlen(line));
	    MLEndPacket(lp);
	    free(line);
	  } else { /* If you don't do this, the kernel hangs around */
	    MLPutString(lp,"exit");
	    MLEndPacket(lp);
	  }	    
	  rl_bind_key('\r',math_return_key);
	  rl_bind_key('\n',math_return_key);
	  MLDisownString(lp,prompt);
	  evaluating = 1;
	} break;
      case BEGINDLGPKT :
      case ENDDLGPKT :
	MLNewPacket(lp); /* Drop Dialog Level */
	break;
      case RESUMEPKT :
        /* We seem to get one of these at startup since I added
           the $ParentLink trick */
	MLNewPacket(lp);  /* Drop it.  What is it anyway? */
        break; 
      case CALLPKT :
      case ENTERTEXTPKT :
      case ENTEREXPRPKT :
      case RETURNEXPRPKT : /* TraditionalForm[Sqrt[x]] */
      case DISPLAYPKT :
      case DISPLAYENDPKT :
      case INPUTSTRPKT :
      case FIRSTUSERPKT :
      case LASTUSERPKT :
      case EVALUATEPKT :
      default:
	fprintf(stderr,"jmath: wierd packet type (%d)\n",packet);
      case SYNTAXPKT :
	read_and_print_expression(lp,stderr);
	fprintf(stderr, "\n");
      }

      if(!line)
	break;
    }

  MLClose( lp); /* this always seems to error in Mathematica 4.0 */
  MLDeinitialize(ml_env);
  printf("\n");
  return 0;
}

/* This should use Unicode, but who knows how? */
static void read_and_print_unquoted_string(MLINK lp,FILE *f,char *extra_string)
{
#if 1
  if(MLGetType(lp) != MLTKSTR)
    fprintf(f,"There should be a string here!!! I'm expecting a string.\n");
#endif
#if 0
  kucharp_ct s;
  long len;
  MLGetByteString( lp, &s,&len,'j');  /* replace 16bit chars with j.  Why?? */
  fwrite(s,1,len,f);
  MLDisownByteString( lp, s, len);
#else
{
  long left,got;
  char buffer[1024];
  do {
    MLGet8BitCharacters(lp, &left, buffer, sizeof(buffer), &got, 'j');
    fwrite(buffer,1,got,f);
    if(interrupt) {
      interrupt = 0;
      fprintf(stderr,"Data Transfer Interrupted\n");
      DebugMLNewPacket(lp); /* dropping what ever is left */
      /* Should longjmp here instead (maybe return) */
      break;
    } 
#if 1
    else 
      /* fprintf(f,"Transfered %ld bytes, %ld left\n",got,left) */;
#endif
  } while(left);
}
#endif
  if(extra_string)
    fprintf(f,extra_string);
}

static void read_and_print_unquoted_symbol(MLINK lp,FILE *f,char *extra_string)
{
  kucharp_ct s;
  long len;
#if 0
  if(MLGetType(lp) != MLTKSYM)
    fprintf(f,"There should be a symbol here!!! I'm expecting a symbol.\n");
#endif
  MLGetByteSymbol( lp, &s,&len,'j');  /* replace 16bit chars with j.  Why?? */
  fwrite(s,1,len,f);
  MLDisownByteSymbol( lp, s, len);
  if(extra_string)
    fprintf(f,extra_string);
}

#if 0
static int read_integer( MLINK lp)
{
  int n;
  if(MLGetType(lp) != MLTKINT)
    fprintf(stderr,"Where is my int!!! I want an int.\n");
  MLGetInteger(lp, &n);
  return n;
}
#endif

static void read_and_print_expression( MLINK lp, FILE *f)
{
        int n;
        long i, len;
        double r;
        static int indent;

        switch( MLGetType( lp)) {
        case MLTKSYM:
#if 0	  
                MLGetSymbol( lp, &s);
                fprintf(f, "%s ", s);
                MLDisownSymbol( lp, s);
#else
		read_and_print_unquoted_symbol(lp,f," "); /*  This should be a symbol */
#endif
                break;
        case MLTKSTR:
#if 0
                MLGetString( lp, &s);
                fprintf(f, "\"%s\" ", s);
                MLDisownString( lp, s);
#else
		fprintf(f,"\"");
		read_and_print_unquoted_string(lp,f,"\"");
#endif
                break;
        case MLTKINT:
                MLGetInteger( lp, &n);
                fprintf(f, "%d ", n);
                break;
        case MLTKREAL:
                MLGetReal( lp, &r);
                fprintf(f, "%g ", r);
                break;
        case MLTKFUNC:
                indent += 3;
                fprintf(f, "\n %*.*s", indent, indent, "");
                if( MLGetArgCount( lp, &len) == 0){
		  error( lp);
                }else{
                        read_and_print_expression( lp, f);
                        fprintf(f, "[");
                        for( i = 1; i <= len; ++i){
                                read_and_print_expression( lp, f);
                                if( i != len) fprintf(f, ", ");
                        }
                        fprintf(f, "]");
                }
                indent -= 3;
                break;
        case MLTKERROR:
        default:
                error( lp);
        }
}

static void
error( lp) MLINK lp;
{
	if (MLError( lp)) {
	       fprintf( stderr, "Error detected by MathLink: %s.\n",
			MLErrorMessage( lp));
	}else{
	       fprintf( stderr, "Error detected by this program.\n");
	}
	MLClose( lp);
	exit( 1);
}

/* Practice in Mathematica itself
l = LinkLaunch["math -mathlink"]
LinkWrite[l, Unevaluated[Plot[x,{x,0,1}]]]
LinkReadyQ[l] && LinkRead[l]
LinkReadyQ[l] && LinkRead[l]
LinkReadyQ[l] && LinkRead[l]

l=LinkLaunch["math -mathlink -runfirst '$SavedParentLink=$ParentLink;$ParentLink=Null' -run '$ParentLink=$SavedParentLink;Remove[$SavedParentLink]'"]
*/
